/******************************************************************************
 *
 * ISO MPEG Audio Subgroup Software Simulation Group (1996)
 * ISO 13818-3 MPEG-2 Audio Decoder - Lower Sampling Frequency Extension
 * 
 * 
 * This software module was originally developed by 
 * 
 *   Matejko Michael (Fraunhofer IIS) 
 * 
 * and edited by 
 * 
 * 
 * in the course of development of the ISO/IEC 13818-3 for reference purposes
 * and its performance may not have been optimized. This software module is
 * an implementation of one or more tools as specified by the ISO/IEC 13818-3.
 * ISO/IEC gives You a royalty-free, worldwide, non-exclusive, copyright
 * license to copy, distribute, and make derivative works of this software
 * module ormodifications thereof for use in implementations of the 
 * ISO/IEC 13818-3  in products that satisfy conformance criteria (if any).
 * Those intending to use this software module in products are advised that 
 * its use may infringe existing patents. ISO/IEC have no liability for use of
 * this software module or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual
 * and image-coding related ITU Recommendations and/or ISO/IEC International
 * Standards. 
 * 
 ******************************************************************************/

/******************************************************************************
 * ATTENTION: Documentation is generated by DoxyGen using "make docs"
 ******************************************************************************/

#include "bitBuffer.h"
#include "debug.h"
#include <stdlib.h>

#define WORD_BITS   (sizeof(unsigned long) * 8)

typedef struct TAG_bitBuffer
{
  unsigned long     *ptr;      /**< Pointer to streambuffer */
  unsigned long     size;      /**< Size of buffer (in bits)*/
  unsigned long     read;      /**< Number of bits read from buffer */
  unsigned long     written;   /**< Number of bits written to buffer */
  unsigned long     underruns; /**< Buffer underrun count */
  FILE              *file;     /**< File handler for input file */
  unsigned long     fileSize;  /**< File size in bytes */
  char              *name;     /**< Name of bit buffer for debuging */
  HANDLE_debug      dbg;       /**< Debug handler */
} bitBuffer_t;


HANDLE_bitBuffer bitBufferInit (unsigned long size, char* name, HANDLE_debug dbg)
{
  unsigned long     size_words = size / WORD_BITS + 1;
  HANDLE_bitBuffer  buf        = NULL;
  
  /* create handler struct */
  if ( ( buf = (HANDLE_bitBuffer) malloc(sizeof(bitBuffer_t)) ) == NULL )
    return NULL; /* MEMORY_ALLOCATION_ERROR */
  
  /* allocate buffer */
  if ( ( buf->ptr = (unsigned long *) malloc(size_words * sizeof(unsigned long)) ) == NULL ) 
  {
    free(buf);
    return NULL; /* MEMORY_ALLOCATION_ERROR */
  }
      
  buf->size = size;
  buf->read = 0;
  buf->written = 0;
  buf->underruns = 0;
  buf->file = NULL;
  buf->name = name;
  buf->dbg = dbg;
    
  return buf;
}


static unsigned long bitBufferRefill(HANDLE_bitBuffer buf)
{
  /* refill only half of buffer, to make it possible to wind buffer back */
  unsigned long freeBytes = bitBufferGetFree(buf) / 8 / 2 ; 
  unsigned long readBytes = 0;
  unsigned char byte;
  
  if (buf->file == NULL)
    return 0;
    
  while(readBytes < freeBytes )
  {
    if ( !fread(&byte, sizeof(unsigned char), 1, buf->file) ) break;
    bitBufferPut(buf, (unsigned long) byte, 8);
    readBytes++;
  }
    
  if (freeBytes > 0)
  {
    /* TODO end of bs */
  }
  return readBytes;
} 


int bitBufferOpenInputFile(HANDLE_bitBuffer buf, char* filename)
{
  fpos_t filePosition;
  
  buf->file = fopen(filename, "rb");
  if (buf->file == NULL) 
  {
    debugPrintf(buf->dbg, D_DEBUG, "%s: Could not find \"%s\"", buf->name, filename);
    return 0;
  }
  
  /* Get file size */
  fseek( buf->file, 0, SEEK_END );
  buf->fileSize = (unsigned long) ftell( buf->file );
  rewind( buf->file );
 
  bitBufferRefill(buf);
  
  return 1;
}


unsigned long bitBufferPut (HANDLE_bitBuffer buf, unsigned long value, unsigned long bits)
{
  unsigned long word_offset, bit_offset, bit, free;
  unsigned long bit_pos = bits;
  
  free = buf->size - buf->written + buf->read;
  
  if (bits > WORD_BITS)
    return 0; /* WRITE_BUFFER_ERROR */
    
  if (bits > free)
  {
    debugPrintf(buf->dbg, D_DEBUG, "%s: bitBufferPut overflow!", buf->name);
    return 0; /* WRITE_BUFFER_ERROR */
  }
  
  while(bit_pos > 0)
  {
    bit_pos--;
    
    bit         = (value & (1 << bit_pos))?1:0;
    word_offset = (buf->written % buf->size) / WORD_BITS;
    bit_offset  = (buf->written % buf->size) % WORD_BITS;
    
    if (bit)
      buf->ptr[word_offset] |= 1 << bit_offset;
    else
      buf->ptr[word_offset] &= ~(1 << bit_offset);
    
    buf->written++;
  }
  
  return bits;
}


unsigned long bitBufferRead (HANDLE_bitBuffer buf, unsigned long *value, unsigned long bits)
{
  unsigned long word_offset, bit_offset, bit, available;
  unsigned long i, refilled;
  *value = 0;
  
  if (bits > WORD_BITS) return 0; /* READ_BUFFER_ERROR */
  
  available = buf->written - buf->read;
  if (bits > available) 
  {
    /* If there is a file associated fill bit buffer with it's content */
    refilled = bitBufferRefill(buf);
    
    if (refilled == 0)
    {
      buf->underruns++;
      if (buf->underruns == 1)
      {
        debugPrintf(buf->dbg, D_DEBUG, "%s: bitBufferRead underrun!", buf->name);
      }
      bits = available;
    }
  }
  
  for(i=0; i<bits; i++)
  {
    word_offset = (buf->read % buf->size) / WORD_BITS;
    bit_offset  = (buf->read % buf->size) % WORD_BITS;
    
    bit = (buf->ptr[word_offset] & (1<<bit_offset))?1:0;

    /* shift value and add new bit */
    *value = (*value << 1) | bit;
    buf->read++;
  }
  
  return bits;
}


unsigned long bitBufferGet (HANDLE_bitBuffer buf, unsigned long bits)
{
  unsigned long value;
  bitBufferRead(buf, &value, bits);
  return value;
}


long bitBufferWind (HANDLE_bitBuffer buf, long bits)
{
  long avail, lost_bits;
  
  if (bits > 0) /* forward */
  {
    avail = buf->written - buf->read;
    if (bits > avail) 
    {
      debugPrintf(buf->dbg, D_DEBUG, "%s: bitBufferWind underrun! Missing %ld bits to fast-forward.", buf->name, bits-avail);
      return 0;
    }
    buf->read += bits;
  }
  else /* bits <= 0 : rewind */
  {
    lost_bits = buf->written - buf->size;
    if (lost_bits < 0) lost_bits = 0; /* buffer not filled to maximum size */
    avail = lost_bits - buf->read;
    if (bits < avail) 
    {
      debugPrintf(buf->dbg, D_DEBUG, "%s: bitBufferWind underrun! Missing %ld bits to wind back.", buf->name, avail-bits);
      return 0;
    }
    buf->read += bits;
  }
  return bits;
}


int bitBufferGoTo (HANDLE_bitBuffer buf, unsigned long pos)
{
  long to_move = pos - buf->read;
  long moved = bitBufferWind(buf, to_move);
  return (to_move == moved);
}


unsigned long bitBufferGetRead (HANDLE_bitBuffer buf)
{
  return buf->read;
}


unsigned long bitBufferGetWritten (HANDLE_bitBuffer buf)
{
  return buf->written;
}


unsigned long bitBufferGetFree (HANDLE_bitBuffer buf)
{
  return buf->size - buf->written + buf->read;
}

unsigned long bitBufferGetAvailable (HANDLE_bitBuffer buf)
{
  unsigned long available = buf->written - buf->read;
  
  if (buf->file)
  {
    unsigned long currentFilePos = (unsigned long) ftell( buf->file);
    unsigned long fileAvailable = buf->fileSize - currentFilePos;
    
    available += fileAvailable*8;
    
    /*if (available > buf->size) available = buf->size;*/ 
  }
  return available;
}

unsigned long bitBufferGetErrors (HANDLE_bitBuffer buf)
{
  return buf->underruns;
}


void bitBufferReset (HANDLE_bitBuffer buf)
{
  buf->read = 0;
  buf->written = 0;
}


unsigned long bitBufferTransfer (HANDLE_bitBuffer in_buf, HANDLE_bitBuffer out_buf, unsigned long bits)
{
  unsigned long words = bits / WORD_BITS;
  unsigned long rest_bits = bits % WORD_BITS;
  unsigned long read_bits;
  unsigned long value;
  unsigned long transferred = 0;
  /* TODO do { ... } while (bits >= 0) */
  
  while (words > 0)
  {
    read_bits    = bitBufferRead(in_buf, &value, WORD_BITS);
    transferred += bitBufferPut(out_buf, value, read_bits);
    words--;
  }
  
  read_bits    = bitBufferRead(in_buf, &value, rest_bits);
  transferred += bitBufferPut(out_buf, value, read_bits);
  
  return transferred;
}


void bitBufferFree (HANDLE_bitBuffer buf)
{
  if (buf) 
  {
    if (buf->file)
      fclose(buf->file);
    free(buf->ptr);
    free(buf);
  }
}
