/******************************************************************************
 *
 * ISO MPEG Audio Subgroup Software Simulation Group (1996)
 * ISO 13818-3 MPEG-2 Audio Decoder - Lower Sampling Frequency Extension
 * 
 * 
 * This software module was originally developed by 
 * 
 *   Matejko Michael (Fraunhofer IIS) 
 * 
 * and edited by 
 * 
 * 
 * 
 * in the course of development of the ISO/IEC 13818-3 for reference purposes
 * and its performance may not have been optimized. This software module is
 * an implementation of one or more tools as specified by the ISO/IEC 13818-3.
 * ISO/IEC gives You a royalty-free, worldwide, non-exclusive, copyright
 * license to copy, distribute, and make derivative works of this software
 * module ormodifications thereof for use in implementations of the 
 * ISO/IEC 13818-3  in products that satisfy conformance criteria (if any).
 * Those intending to use this software module in products are advised that 
 * its use may infringe existing patents. ISO/IEC have no liability for use of
 * this software module or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual
 * and image-coding related ITU Recommendations and/or ISO/IEC International
 * Standards. 
 * 
 ******************************************************************************/

/******************************************************************************
 * ATTENTION: Documentation is generated by DoxyGen using "make docs"
 ******************************************************************************/

#include "frameData.h"
#include "bitBuffer.h"
#include "common.h"

/******************************************************************************
 * private
 ******************************************************************************/

static const
char *version_names[VERSION_COUNT] = 
{ "MPEG-1", "MPEG-2 LSF",
#ifdef MPEG_2_5
"reserved","MPEG 2.5"
#endif
};

static const
char *layer_names[3] = { "Layer I", "Layer II", "Layer III" };

static const
char *mode_names[5] = { "stereo", "j-stereo", "dual-ch", "single-ch" , "multi-ch"};

static const
double s_freq[VERSION_COUNT][4] =
{
  /* MPEG-1 */
  {44.1, 48, 32, 0},
  /* MPEG-2 */
  {22.05, 24, 16, 0},
#ifdef MPEG_2_5
  /* reserved */
  {0,0,0,0},
  /* MPEG-2.5 */
  {11.025, 12, 8, 0}
#endif
};


static const
int bitrate[VERSION_COUNT][3][15] = 
{
  { /* MPEG-1 */
    {0,32,64,96,128,160,192,224,256,288,320,352,384,416,448},
    {0,32,48,56,64,80,96,112,128,160,192,224,256,320,384},
    {0,32,40,48,56,64,80,96,112,128,160,192,224,256,320}
  },
  { /* MPEG-2 */
    {0,32,48,56,64,80,96,112,128,144,160,176,192,224,256},
    {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160},
    {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160}
  },
#ifdef MPEG_2_5
  { /* reserved */
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}
  },
  { /* MPEG-2.5 */
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
    {0,8,16,24,32,40,48,56,64,0,0,0,0,0,0}
  }
#endif
};


/******************************************************************************
 * public functions
 ******************************************************************************/

/**
 * get header information
 * This function needs the bitBuffer buf to be at the first bit of 
 * syncword.
 */
int getFrameData(frameData *fd, HANDLE_bitBuffer buf)
{
  /* skip syncword */
  int sync = bitBufferGet(buf,11);
  
  if (sync != 0x7FF) return 0;

  /* get header content from bit buffer */
  fd->version            = 3-bitBufferGet(buf,2);
  fd->lay                = 4-bitBufferGet(buf,2);
  fd->error_protection   = !bitBufferGet(buf,1); /* error protection bit TRUE/FALSE */
  fd->bitrate_index      = bitBufferGet(buf,4); /* bitrate index */
  fd->sampling_frequency = bitBufferGet(buf,2); /* sampling rate index */
  fd->padding            = bitBufferGet(buf,1); /* padding bit */
  fd->extension          = bitBufferGet(buf,1); /* private bit */
  fd->mode               = bitBufferGet(buf,2); /* channel mode */ 
  fd->mode_ext           = bitBufferGet(buf,2); /* mode extension */
  fd->copyright          = bitBufferGet(buf,1); /* copyright bit */
  fd->original           = bitBufferGet(buf,1); /* original bit */
  fd->emphasis           = bitBufferGet(buf,2); /* emphasis */
  
  /* check version */
#ifndef MPEG_2_5
  /* only MPEG1 or MPEG2 allowed */
  if (fd->version != MPEG_PHASE1 && fd->version != MPEG_PHASE2) return 0;
#else
  /* every possible combination except MPEG_RSVD_VER allowed */
  if (fd->version == MPEG_RSVD_VER) return 0;
#endif
  /* check layer */
  if (fd->lay > 3) return 0;
    
  /* check bitrate index */
  if (fd->bitrate_index == 0xF) return 0;
  if (fd->version == MPEG_PHASE2_5 && fd->bitrate_index > 8) return 0;
  
  /* check sampling freq index */
  if (fd->sampling_frequency == 0x4) return 0;
  
  /* get crc value */
  if (fd->error_protection)
  {
    fd->crc              = bitBufferGet(buf, 16);
  }
  
  if (fd->lay == LAYER3)
  {
    int mdb_size = (fd->version==1)?9:8;
    fd->main_data_begin  = bitBufferGet(buf,mdb_size);
    bitBufferWind(buf, -mdb_size);  /* main_data_begin is part of side info */
  }
  
  
  switch (fd->lay)
  {
    case LAYER1:
      fd->samples = 384;
      break;
    case LAYER2:
      fd->samples = 1152;
      break;
    case LAYER3:
      fd->samples = (fd->version == MPEG_PHASE1)?1152:576;
      break;
  }

  fd->sfreq       = s_freq[fd->version][fd->sampling_frequency];
  fd->bitrate     = bitrate[fd->version][fd->lay-1][fd->bitrate_index];
  fd->channels    = (fd->mode == MPG_MD_MONO)?MONO:STEREO;
  fd->slotbits    = (fd->lay==LAYER1)?32:8;
  
  fd->strVersion  = version_names[fd->version];
  fd->strLayer    = layer_names[fd->lay-1];
  fd->strMode     = mode_names[fd->mode];

  /* TODO recognize bad headers and return 0 */
  return 1;  
}
