/******************************************************************************
 *
 * ISO MPEG Audio Subgroup Software Simulation Group (1996)
 * ISO 13818-3 MPEG-2 Audio Decoder - Lower Sampling Frequency Extension
 * 
 * 
 * This software module was originally developed by 
 * 
 *   Matejko Michael (Fraunhofer IIS) 
 * 
 * and edited by 
 * 
 * 
 * 
 * in the course of development of the ISO/IEC 13818-3 for reference purposes
 * and its performance may not have been optimized. This software module is
 * an implementation of one or more tools as specified by the ISO/IEC 13818-3.
 * ISO/IEC gives You a royalty-free, worldwide, non-exclusive, copyright
 * license to copy, distribute, and make derivative works of this software
 * module ormodifications thereof for use in implementations of the 
 * ISO/IEC 13818-3  in products that satisfy conformance criteria (if any).
 * Those intending to use this software module in products are advised that 
 * its use may infringe existing patents. ISO/IEC have no liability for use of
 * this software module or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual
 * and image-coding related ITU Recommendations and/or ISO/IEC International
 * Standards. 
 * 
 ******************************************************************************/

/******************************************************************************
 * ATTENTION: Documentation is generated by DoxyGen using "make docs"
 ******************************************************************************/

#include "frameDeMux.h"
#include "frameData.h"
#include <stdlib.h>
#include "debug.h"


#define   SYNC_WORD         0xFFE00000
/* mask to sync complete header */
#define   SYNC_MASK         0xFFFE0000

/* mask and value to compare if free mode active */
#define   FREE_MASK         0x0000F000
#define   NO_SYNC_FOUND     (unsigned) -1
#define   HEADER_SIZE       32
#define   CRC_WORD_SIZE     16

/* max(main_data_pointer) * 8 = 2^9 * 8 = 4096 */
#define   MAX_MAIN_DATA 68640

typedef struct TAG_frameDeMux
{
  HANDLE_debug      debug;              /* debug handler */
  HANDLE_bitBuffer  bitstream;          /* input bitstream buffer */
  HANDLE_bitBuffer  mainData;           /* mainDataBuffer for layer3 */
  unsigned long     raw_header;         /* 32 bit header used by syncroutine */
  unsigned long     freeFrameSlots;     /* static frame size for free format */
  unsigned long     frameCount;         /* current frame number */
  unsigned long     start;              /* begining of next frame */
  unsigned long     length;             /* length of next frame */
  frameData         frameData;          /* frame data of next frame */
  frameConfig       frameConfig;        /* configuration of last demuxed frame */
  int               frameValid;         /* is the current frame valid? */
} frameDeMux_t;

/* 
 *  seek_sync_header(buf, header)
 *
 *  Syncronize to bitstream
 *  If *header == 0 seek only for syncword and get 32-bit header (this is
 *  used for first syncing to bitstream.)
 *  If *header != 0 syncronize to bitstream using known data from header
 *  (it is not sufficient to simply search for the 11 bit syncword because
 *  this pattern is often present in the bit stream.)
 *
 *  The function seeks for a byte aligned sync word in the bitstream and
 *  places the bitstream pointer right after the sync.
 *  It returns the sync position (beginning of syncword) if the sync was 
 *  found otherwise it returns NO_SYNC_FOUND.
 *
 *  This is not without problems, as
 *  1.  this pattern may also be emulated by the bit stream
 *  2.  it is not excluded by the standard to change the crc protection 
 *      from frame to frame.
 * 
 *  In free mode the frame size is determined by the distance between two
 *  syncwords. In that case do not print warnings! 
 */

/**
 * Seek for next syncword in bitstream
 */
static 
unsigned long seekSync (HANDLE_frameDeMux fdm)
{
  HANDLE_bitBuffer buf   = fdm->bitstream;
  unsigned long read;
  unsigned long align    = (8-(bitBufferGetRead(buf)%8))%8;
  unsigned long skipped  = 0; /* count skipped bits for warning */
  
  /* check if enough available data */
  if (bitBufferGetAvailable(buf) < HEADER_SIZE+align) return NO_SYNC_FOUND;
  
  /* align at first */
  bitBufferGet(buf, align);
  
  /* read 32 bits from bitstream */
  read = bitBufferGet(buf, HEADER_SIZE);
  
  while (bitBufferGetAvailable(buf) >= 8)
  {
    /* check syncword */
    if ( (read & SYNC_WORD) == SYNC_WORD )
    {
      /* syncword found */
      bitBufferWind(buf, -(HEADER_SIZE));
      debugPrintf(fdm->debug, D_DEBUG, "Sync found at: %X (skipped %lu bits)", bitBufferGetRead(buf)/8, skipped);
      return bitBufferGetRead(buf);
    }
      
    /* no syncword found, seek 1 byte foreward */
    read <<= 8;
    read |= bitBufferGet(buf, 8);
    skipped += 8;
  }
  
  /* no more data available in bit buffer */
  return NO_SYNC_FOUND;
}

static
unsigned int getFrameLength (HANDLE_frameDeMux fdm)
{
  HANDLE_bitBuffer buf = fdm->bitstream;
  frameData *fd        = &fdm->frameData;
  double bitsPerSlot   = fd->slotbits;
  int slots; /* frame size in slots */
  
  /* get frame size */
  if (fd->bitrate == 0) 
  {
    /* frame size for FREE FORMAT */
    if (fdm->freeFrameSlots == 0)
    {
      unsigned long header_1 = 0x0;
      unsigned long header_2 = 0x0;
      unsigned long next_sync_pos = 0;

      /* get first header */
      bitBufferGoTo(buf, fdm->start);
      header_1 = bitBufferGet(buf, HEADER_SIZE);
      
      /* search next header */
      while ((header_1 & SYNC_MASK) != (header_2 & SYNC_MASK))
      {
        next_sync_pos = seekSync(fdm); /* seek next sync */
        if (next_sync_pos == NO_SYNC_FOUND)
        {
           debugPrintf(fdm->debug, D_ERROR, "Free format indicated, but no second frame found.");
           return 0;
        }
        header_2 = bitBufferGet(buf, HEADER_SIZE);
      }

      /* go to beginning of sync-search */ 
      bitBufferGoTo(buf, fdm->start);

      fdm->freeFrameSlots = (next_sync_pos - fdm->start) / fd->slotbits;
      debugPrintf(fdm->debug, D_NOTICE_L1, "Free format indicated with %d slots per frame.", fdm->freeFrameSlots);
    }
    slots = fdm->freeFrameSlots;
  }    
  else
  {
    /* frame size calculation */
    double samplesPerFrame    = fd->samples;
    double samplingFrequency  = fd->sfreq;
    double bitRate            = fd->bitrate;
    
    /* Calc no of slots used for one bitstream frame. 
     * It is necessary to divide by bitsPerSlot to get right values (caused by rounding) */ 
    slots = (int)( (samplesPerFrame * bitRate)  / samplingFrequency / bitsPerSlot );
  }
  
  if (fd->padding) slots += 1;
  
  return slots * bitsPerSlot;  
}

static
unsigned int getSideInfoSize (frameData *frame)
{
  unsigned int size;
  int nch   = frame->channels;
  int mpeg1 = (frame->version==MPEG_PHASE1);
  int mode3 = (frame->mode==MPG_MD_MONO);
  
  /* side info bits               [ mpeg1   | mpeg2, 2.5 ]
     -----------------------------------------------------
        main_data_begin             9         8
        if (mode3)
          private_bits              5         1
        else
          private_bits              3         2
        scfsi                       0         1*nch*4
        
        (
          part2_3_length            12        12
          big_values                9         9
          global_gain               8         8
          scalfac_compress          4         9
          window_switching_flag     1         1
          if (window_switching_flag)
          {
            transfer block_type     2         2
            mixed_block_flag        1         1
            table_select            5*2       5*2
            subblock_gain           3*3       3*3
          }                       = 22      = 22
          else
          {
            table_select            5*3       5*3
            region0_count           4         4
            region1_count           3         3
          }                       = 22      = 22
          preflag                   0         1
          scalefac_scale            1         1
          couttable_slect           1         1
        )                         = 58        64
        * ngr                       2         1
                                  = 116       64
        * nch
  */
  
  if (mpeg1)
  {
    size  = 9;
    size += (mode3)? 5 : 3;
    size += 4 * nch;
    size += 59 * 2 * nch ;  /* gr = 2 */
  }
  else /* mpeg2, 2.5 */
  { 
    size  = 8;
    size += (mode3)? 1 : 2;
    size += 63 * nch ; /* gr = 1 */
  }
  
  return size;
}


static
ERROR_CODE_frameDeMux checkCurrentFrame(HANDLE_frameDeMux fdm)
{
  unsigned long expected_pos;
  unsigned long current_pos;
  unsigned long available, expected;

  /* check current bitstream position */
  expected_pos = fdm->start;
  current_pos  = bitBufferGetRead(fdm->bitstream);
  if (current_pos != expected_pos)
  {
    debugPrintf(fdm->debug, D_ERROR, "Internal error: bitstream isn't at right position! Difference: %ld bits.", (expected_pos - current_pos));
    debugLeave(fdm->debug);
    return FMUX_UNKNOWN_ERROR;
  } 
  
  /* check available data in bitstream */
  available = bitBufferGetAvailable(fdm->bitstream);
  expected  = fdm->length;
  if (expected == 0)
  {
    debugPrintf(fdm->debug, D_NOTICE_L2, "End of bitstream.");
    return FMUX_END_OF_STREAM;
  }
  else if (available < expected)
  {
    debugPrintf(fdm->debug, D_NOTICE_L2, "Not enough data at end of stream. Expected %lu bits, but %lu bits available.", expected, available);
    return FMUX_END_OF_STREAM;
  }

  fdm->frameConfig.length     = fdm->length;
  fdm->frameConfig.samplerate = fdm->frameData.sfreq * 1000;
  fdm->frameConfig.bitrate    = fdm->frameData.bitrate;
  fdm->frameConfig.channels   = fdm->frameData.channels;
  fdm->frameConfig.version    = fdm->frameData.version + 1;
  fdm->frameConfig.layer      = fdm->frameData.lay;
  
  if (fdm->frameConfig.bitrate == 0)
  {
    fdm->frameConfig.bitrate = fdm->frameConfig.samplerate / fdm->frameData.samples * fdm->length / 1000;
  }  
  
  return FMUX_NO_ERROR;
}


static
ERROR_CODE_frameDeMux checkNextFrame(HANDLE_frameDeMux fdm)
{
  ERROR_CODE_frameDeMux ret;
  unsigned long current_pos;

  /* next frame start position */
  fdm->start = fdm->start + fdm->length;
  fdm->length = 0;  
  
  /* check current bitstream position */
  current_pos = bitBufferGetRead(fdm->bitstream);
  if (current_pos != fdm->start)
  {
    /* IMPOSSIBLE current_pos must be at expected_sync_pos! */
    debugPrintf(fdm->debug, D_ERROR, "Internal error: bitstream isn't at right position! Difference: %ld bits.", (fdm->start - current_pos));
    return FMUX_UNKNOWN_ERROR;
  }
  
  /* check if more data available */
  if (bitBufferGetAvailable(fdm->bitstream) < HEADER_SIZE)
  {
    debugPrintf(fdm->debug, D_NOTICE_L2, "No more data in bitstream.");
    /* No next header, current frame is valid */
    return FMUX_NO_ERROR;
  }

  /* check next frame header */
  ret = FMUX_NO_ERROR;
  while (!getFrameData(&fdm->frameData, fdm->bitstream))
  {
    /* current frame does not ent with a valid new header -> do resync */
    unsigned long sync;
    
    debugPrintf(fdm->debug, D_NOTICE_L2, "Next frame header invalid. Searching new sync.");

    bitBufferReset(fdm->mainData); /* if an error occures the mainData is invalid! */
    
    /* back to frame start and move one bit foreward */
    bitBufferGoTo(fdm->bitstream, fdm->start - fdm->length + 1);

    /* search next sync */
    sync = seekSync(fdm);
    if (sync == NO_SYNC_FOUND)
    {
      current_pos = bitBufferGetRead(fdm->bitstream);
      debugPrintf(fdm->debug, D_WARNING, "Sync not found. Skipped %lu bits at end of bitstream.", current_pos-fdm->start);
      fdm->start = current_pos;
      return FMUX_INVALID_FRAME_LENGTH;
    }
    /* new frame starts at new sync position */
    fdm->start = sync;
    
    ret = FMUX_INVALID_FRAME_LENGTH; /* current frame is invalid */
  }

  bitBufferGoTo(fdm->bitstream, fdm->start);
  fdm->length = getFrameLength(fdm);
  return ret;
  
}


static
int seekFirstFrame(HANDLE_frameDeMux fdm)
{
  frameData frameDataNext;
  
  while (1)
  {
    fdm->start = seekSync(fdm);
    if (fdm->start == NO_SYNC_FOUND)
    {
      debugPrintf(fdm->debug, D_DEBUG, "No sync found.");
      return 0;
    }

    if (getFrameData(&fdm->frameData, fdm->bitstream))
    {
      /* first header valid, check next header */
      fdm->length = getFrameLength(fdm);
      bitBufferGoTo(fdm->bitstream, fdm->start + fdm->length);
      if (getFrameData(&frameDataNext, fdm->bitstream))
      {
        /* next header also valid, frame is valid */
       bitBufferGoTo(fdm->bitstream, fdm->start);
       return 1; 
      }
    }

    /* invalid header, search next sync */
    bitBufferGoTo(fdm->bitstream, fdm->start + 1);
    
  }
}


/**************************************************************************

  Initialize frameMux

*************************************************************************/
HANDLE_frameDeMux frameDeMuxInit(HANDLE_bitBuffer bitstream, HANDLE_debug debug)
{
  HANDLE_frameDeMux fdm;

  debugEnter(debug, "frameDeMuxInit()");
  
  /* allocate handler struct */
  fdm = (HANDLE_frameDeMux) malloc(sizeof(frameDeMux_t));
  if ( fdm == NULL )
  {
    debugPrintf(debug, D_ERROR, "Could not allocate memory for Handler.");
    debugLeave(debug);
    return NULL;
  }
    
  /* initialize buffer for mp3 maindata */  
  fdm->mainData = bitBufferInit(MAX_MAIN_DATA, "mainData", debug);
  if ( fdm->mainData == NULL ) 
  {
    debugPrintf(debug, D_ERROR, "Could not initialize main data buffer.");
    debugLeave(debug);
    return NULL;
  }
  
#if 0
  fdm->bitstream = bitBufferInit(MAX_BS_FRAME_SIZE, "bitstreamBuffer", debug);
  if ( fdm->bitstream == NULL ) 
  {
    debugPrintf (dbg, D_ERROR, "Could not initialize bitstream buffer.");
    debugLeave(dbg);
    return NULL;
  }
  if ( !bitBufferOpenInputFile(fdm->bitstream, filename) ) 
  {
    debugPrintf (dbg, D_ERROR, "Could not open input file.");
    debugLeave(dbg);
    return NULL;
  }
#endif
  
  fdm->bitstream = bitstream;
  fdm->debug = debug;
  fdm->raw_header = 0;
  fdm->freeFrameSlots = 0;
  fdm->frameCount = 0;
  fdm->start = 0;
  fdm->length = 0;
  
  /* Sync first frame */
  if (!seekFirstFrame(fdm))
  {
    debugPrintf(fdm->debug, D_ERROR, "Invalid bitstream.");
    frameDeMuxFree(fdm);
    debugLeave(debug);
    return NULL;
  }

  debugPrintf(fdm->debug, D_NOTICE_L2, "First sync found at position %ld (%X)", fdm->start, fdm->start);
  debugLeave(debug);
  return fdm;
}


/**************************************************************************

  demux next frame in bitstream and output it into audioFrame buffer

*************************************************************************/
ERROR_CODE_frameDeMux frameDeMuxNext(HANDLE_frameDeMux fdm, HANDLE_bitBuffer audioFrame)
{
  ERROR_CODE_frameDeMux ret;
  ERROR_CODE_frameDeMux c_ret;
  
  debugEnter(fdm->debug, "frameDeMuxNext()");
  
  /* clean audio frame */
  bitBufferReset(audioFrame);

  /* set current frame as invalid */ 
  fdm->frameValid = 0;

  ret = checkCurrentFrame(fdm);
  if (ret != FMUX_NO_ERROR)
  {
    debugLeave(fdm->debug);
    return ret;
  }
  
  debugPrintf(fdm->debug, D_NOTICE_L2, "DeMuxing frame %ld at position %ld (%X)", fdm->frameCount, fdm->start, fdm->start);

  /* get audio data by layer */
  switch (fdm->frameData.lay)
  {
    case LAYER1:
    case LAYER2:
      /* Rewind bitstream buffer because header was read before */
      /* for layer1 and layer2: audioFrame and bitstreamFrame are the same, just copy it */
      bitBufferTransfer(fdm->bitstream, audioFrame, fdm->length);
      break;
      
    case LAYER3:
      {
        /* for layer3: the mainData of an audioFrame may be part of previous bitstreamFrames */
        unsigned long main_data_begin;
        unsigned long size_si, size_mdb, size_md, md_free, crc_size;
        long md_tomove, md_moved, md_avail;
        crc_size = (fdm->frameData.error_protection)?CRC_WORD_SIZE:0;
        
        /* Transfer header and CRC word into audio frame buffer */
        bitBufferTransfer(fdm->bitstream, audioFrame, HEADER_SIZE + crc_size);
        
        /* Get side info and main_data_begin size */
        size_si  = getSideInfoSize(&fdm->frameData);
        size_mdb = (fdm->frameData.version==MPEG_PHASE1)?9:8;
        
        /* Get pointer to beginning of mainData for this frame */
        main_data_begin = bitBufferGet(fdm->bitstream, size_mdb);
        debugPrintf(fdm->debug, D_NOTICE_L2, "Main data pointer: %d", main_data_begin);
        
        /* Write it into the audioFrame */
        bitBufferPut(audioFrame, main_data_begin, size_mdb);
  
        /* Transfer rest of side info from bitstream to audioFrame */
        bitBufferTransfer(fdm->bitstream, audioFrame, (size_si-size_mdb) );
        
        /* Refill mainDataBuffer with rest of frame data from bitstream */
        size_md = (fdm->length - size_si - HEADER_SIZE - crc_size);
        md_free = bitBufferGetFree(fdm->mainData);
        if (size_md > md_free)
        {
          debugPrintf(fdm->debug, D_WARNING, "Not enough space in main data buffer. Skipping %ld bits of recent main data.", size_md-md_free);
          bitBufferWind(fdm->mainData, size_md-md_free);
        }
        bitBufferTransfer(fdm->bitstream, fdm->mainData, size_md);
        debugPrintf(fdm->debug, D_NOTICE_L2, "Filled main data buffer with %ld bits. Current fill level is %.1f%%", size_md, (float)bitBufferGetAvailable(fdm->mainData)/(MAX_MAIN_DATA)*100 );
       
        /* Move read pointer to beginning of main data of current frame.
         * this may be data from last bitstream frame! */
        md_tomove = -(8*main_data_begin);
        md_moved = bitBufferWind(fdm->mainData, md_tomove);
        
        /* Check if enough main data was available in buffer to resolve backpointer */
        if (md_moved != md_tomove)
        {
          debugPrintf(fdm->debug, D_NOTICE_L2, "Not enough prebuffered main data to demux frame %d.", fdm->frameCount);
          bitBufferReset(audioFrame);
          /* Skip current main data: fast-forward main data buffer to end */
          bitBufferWind(fdm->mainData, bitBufferGetAvailable(fdm->mainData));
          ret = FMUX_NOT_ENOUGH_DATA;
        }
        else
        {
          /* Transfer main data from mainDataBuffer to audioFrame */
          md_avail = bitBufferGetAvailable(fdm->mainData);
          bitBufferTransfer(fdm->mainData, audioFrame, md_avail);
        }
      }
      break; /* LAYER3 */
  }
  
  c_ret = checkNextFrame(fdm);
  if (c_ret != FMUX_NO_ERROR)
  {
    bitBufferReset(audioFrame);
    debugLeave(fdm->debug);
    return c_ret;
  }
  
  /* set current frame as valid */ 
  if (ret == FMUX_NO_ERROR)
  {
    fdm->frameValid = 1;
  }
  fdm->frameCount++;
  
  debugPrintf(fdm->debug, D_NOTICE_L2, "Frame with length %ld demuxed.", fdm->length);
  debugLeave(fdm->debug);
  return ret;
}


/**
 * Get info about total amount of processed bits
 */
unsigned long frameDeMuxGotBits(HANDLE_frameDeMux fdm)
{
  return bitBufferGetRead(fdm->bitstream);
}

/**
 * Get info about last demuxed frame
 */
frameConfig frameDeMuxGetCurrentConfig(HANDLE_frameDeMux fdm)
{
  if (!fdm->frameValid)
  {
    debugPrintf(fdm->debug, D_WARNING, "Trying to read config of invalid frame.");
    fdm->frameConfig.length     = 0;
    fdm->frameConfig.samplerate = 0;
    fdm->frameConfig.bitrate    = 0;
    fdm->frameConfig.channels   = 0;
    fdm->frameConfig.version    = 0;
    fdm->frameConfig.layer      = 0;
  }
  return fdm->frameConfig;
}


/**
 * Free 
 */
void frameDeMuxFree(HANDLE_frameDeMux fdm)
{
  if (fdm)
  {
    bitBufferFree(fdm->mainData);
    free(fdm);
  }
}


