/******************************************************************************
 *
 * ISO MPEG Audio Subgroup Software Simulation Group (1996)
 * ISO 13818-3 MPEG-2 Audio Decoder - Lower Sampling Frequency Extension
 * 
 * 
 * This software module was originally developed by 
 * 
 *   Matejko Michael (Fraunhofer IIS) 
 * 
 * and edited by 
 * 
 * 
 * 
 * in the course of development of the ISO/IEC 13818-3 for reference purposes
 * and its performance may not have been optimized. This software module is
 * an implementation of one or more tools as specified by the ISO/IEC 13818-3.
 * ISO/IEC gives You a royalty-free, worldwide, non-exclusive, copyright
 * license to copy, distribute, and make derivative works of this software
 * module ormodifications thereof for use in implementations of the 
 * ISO/IEC 13818-3  in products that satisfy conformance criteria (if any).
 * Those intending to use this software module in products are advised that 
 * its use may infringe existing patents. ISO/IEC have no liability for use of
 * this software module or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual
 * and image-coding related ITU Recommendations and/or ISO/IEC International
 * Standards. 
 * 
 ******************************************************************************/

/******************************************************************************
 * ATTENTION: Documentation is generated by DoxyGen using "make docs"
 ******************************************************************************/

/** @mainpage mpeg12dec Documentation
 * 
 * @section sec_intro Introduction
 *
 * This is the introduction.
 *
 * @section sec_compile Compiling
 *
 * @code
 * make [USE_AFSP=1] [MPEG_2_5=1]
 * @endcode
 * When USE_AFSP is definded the program will be linked with the AFsp library "libtsp".
 * To enable MPEG 2.5 decoding just define the MPEG_2_5 variable as shown.
 * 
 * 
 * @section sec_run Running the decoder
 * @code
 * mpeg12dec [options] <infile> [<outfile>]
 * @endcode
 * 
 * For option details just run
 * @code
 * mpeg12dec
 * @endcode
 * 
 * @section sec_struct Structure of the decoder
 * 
 * The decoder is splitted into three parts:
 * 
 * \li @ref mpeg12dec.c "Decoder front end (mpeg12dec.c)" \n
 *  This contains the main() function. And calls the other two parts.
 * \li @ref frameDeMux.c "Frame demuxer (frameDeMux.c)" \n
 *  This module demuxes the physical MPEG frames from the bitstream into logical
 *  audio frames which are needed by the decoder core. For layer 1 and 2 
 *  pysical bistream frames are identical to logical audio frames. For layer3
 *  the main data pointer is resolved to create audio frames from bitstream
 *  frames. 
 * \li @ref decLayer123.c "Decoder core (decLayer123.c)"
 *
 *   Note that the classes in the above graph are clickable
 *   (in the HTML output)
 * 
 */



/******************************************************************************
 * includes
 ******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "decLayer123.h"
#include "frameDeMux.h"
#include "fileWriter.h"
#include "debug.h"

/******************************************************************************
 * type definitions
 ******************************************************************************/

#define REM( text ) debugPrintf(debug, D_DEBUG_L2, text);

#ifndef FILENAME_MAX
#define FILENAME_MAX 255
#endif

#define MAX_FRAME_PCM_SAMPLES 13824 /* SSLIMIT * SBLIMIT * STEREO * SCALE_BLOCK */
/* (((max_framesize * max_bitrate) / min_sfreq) + max_padding ) = (((1152 * 448) / 8)  + 32) = 8068 bytes */
#define MAX_AU_FRAME_SIZE 64544 /* bits */
#define MAX_BS_FRAME_SIZE 64544 /* bits */ 

typedef struct 
{
  char *decPara;
  char *encoded_file_name;
  char *decoded_file_name;
  FORMAT_fileWriter format;
  errorLevel bs_dbg;
} arguments;


typedef enum 
{
  NO_ERROR = 0,
  UNKNOWN_ERROR,
  END_OF_STREAM,
  FILE_OPEN_ERROR,          /* error opening file */
  MEMORY_ALLOCATION_ERROR   /* error allocating memory */
} ERROR_CODE;


/******************************************************************************
 * functions
 ******************************************************************************/

/**
 * Print usage help on stdout
 * 
 * @param programName Should be argument argv[0]
 */
static void usage(char* programName)
{
  printf("usage: %s [options] <infile> [<outfile>]\n", programName);
  printf("options:\n");
  printf(" -A        output is AIFF PCM\n");
  printf(" -R        output is RAW  PCM (big endian)");
#ifndef USELIB_AFsp
  printf(" (default)\n");
#else
  printf("\n");
  printf(" -W        output is WAVE PCM (default)\n");
#endif
  printf(" -C        24 bit output (default is 16 bit)\n");
  printf(" -n <ch>   force number of channels\n");
  printf("            1   - mono\n");
  printf("            2   - stereo\n");
  printf(" -d <lev>  debug level (default 2)\n");
  printf("            0   - Errors only\n");
  printf("            1   - Warnings\n");
  printf("            2-5 - Notices (for bitstream debuging)\n");
  printf("            6-9 - Debugging (for program debuging)\n");
  printf(" -l <str>  MPEG 1/2 decoder library parameter string\n");
  printf(" <infile>  input bitstream of encoded audio\n");
  printf(" <outfile> output PCM sound file (dflt <infile>.dec)\n\n");
}





/**
 * Parse command line
 */
static int getArguments(
    int argc, char **argv,    /**< standard argc, argv combination */
    arguments *args,          /**< arguments struct that will be filled */
    HANDLE_debug debug        /**< debug handler */
)
{
  int i = 0;
  int set_debug    = 0;
  int set_dec_para = 0;
  int set_channels = 0;
  int set_int24    = 0;
  int set_format   = 0;
  int set_file     = 0;  
  
  debugEnter(debug, "getArguments()");

  /* init values */
  args->format = 0;
  args->bs_dbg = D_NOTICE;
  args->decPara = "";

  /* for each argument */
  for (i = 1; i < argc; i++)
  {
    char *arg  = argv[i];
    char *next = (i+1<argc)? argv[i+1] : "";
    
    if (*arg == '-')
    {
      switch(*(++arg))
      {
        
        case 'd':
          if (set_debug++ != 0)
          {
            debugPrintf (debug, D_ERROR, "Multiple debug option (-d).");
            debugLeave (debug);
            return 0;
          }
          else
          {
            int dbg_lev = atoi(next);
            if(dbg_lev < 0 || dbg_lev > 9) 
            {
              debugPrintf (debug, D_ERROR, "-d %s: not between 0 and 9", next);
              debugLeave (debug);
              return 0;
            }
            debugSetLevel(debug, dbg_lev);
            debugPrintf (debug, D_DEBUG, "Setting debug level to: %d", dbg_lev);
          }
          i++; /* skip next - used as argument */
          continue;

          
        case 'n':
          if (set_channels++ != 0)
          {
            debugPrintf (debug, D_ERROR, "Multiple channel set option (-n).");
            debugLeave (debug);
            return 0;
          }
          switch(atoi(next))
          {
            case 1:
              debugPrintf (debug, D_NOTICE_L3, "Forcing mono option used.");
              args->format += FW_FORCE_MONO;
              break;
            case 2:
              debugPrintf (debug, D_NOTICE_L3, "Forcing stereo option used.");
              args->format += FW_FORCE_STEREO;
              break;
            default:
              debugPrintf (debug, D_ERROR, "-n %s: must be 1 or 2", next);
              debugLeave (debug);
              return 0;
          }
          i++; /* skip next - used as argument */
          continue;

          
        case 'l':
          if (set_dec_para++ != 0)
          {
            debugPrintf (debug, D_ERROR, "Multiple decoder parameters option (-l).");
            debugLeave (debug);
            return 0;
          }
          else
          args->decPara = next;
          debugPrintf (debug, D_NOTICE_L3, "Decoder parameter string: %s", next);
          i++;
          continue; 


        case 'C':
          if (set_int24++ != 0)
          {
            debugPrintf (debug, D_ERROR, "Multiple 24 bit output option (-C).");
            debugLeave (debug);
            return 0;
          }
          debugPrintf (debug, D_DEBUG, "Selected 24 bit output format.");
          args->format += FW_INT24;
          continue;

          
        case 'A':
          debugPrintf (debug, D_NOTICE_L3, "Selected AIFF format for output.");
          set_format++;
          args->format += FW_AIFF;
          continue;


        case 'E':
          debugPrintf (debug, D_NOTICE_L3, "Selected ESPS format for output.");
          set_format++;
          args->format += FW_ESPS;
          continue;


        case 'R':
          debugPrintf (debug, D_NOTICE_L3, "Selected RAW format for output.");
          set_format++;
          args->format += FW_RAW;
          continue;


#ifdef USELIB_AFsp
        case 'W':
          debugPrintf (debug, D_NOTICE_L3, "Selected WAVE format for output.");
          set_format++;
          args->format += FW_WAVE;
          continue;
#endif


        default:
          debugPrintf (debug, D_ERROR, "Unrecognized option: %s", arg);
          debugLeave (debug);
          return 0;
          
      } /* switch */
    } /* if */
    else
    {
      /* argument not beginning with "-" */
      switch(++set_file)
      {
        case 1:
          args->encoded_file_name = arg;
          continue;
          
        case 2:
          args->decoded_file_name = arg;
          continue;
        
        default:
          debugPrintf (debug, D_ERROR, "Unexpected parameter: %s", arg);
          debugLeave (debug);
          return 0;
      } /* switch */
    } /* if */
  } /* for */
      
  if (set_format > 1)
  {
    debugPrintf (debug, D_ERROR, "Multiple format select.");
    debugLeave (debug);
    return 0;
  }
  else if (set_format == 0)
  {
#ifdef USELIB_AFsp
    args->format += FW_WAVE;
#else
    args->format += FW_RAW;
#endif
  }
    
  if (set_file == 0)
  {
    debugPrintf (debug, D_ERROR, "No input file specified.");
    usage(argv[0]);
    decLayer123Info(stdout); /* Layer3 EDIT */
    debugLeave (debug);
    return 0;
  }

  if (set_file == 1)
  {
    static char dec_file[FILENAME_MAX]; /* memory for filename */
    args->decoded_file_name = dec_file;
    strcpy(args->decoded_file_name, args->encoded_file_name);
    strcat(args->decoded_file_name, ".dec");
  }    


  /* report results of dialog / command line */
  debugPrintf (debug, D_NOTICE_L1, "Input file:  %s", args->encoded_file_name);
  debugPrintf (debug, D_NOTICE_L1, "Output file: %s", args->decoded_file_name);
  
  debugLeave(debug);
  return 1;
}



/**
 * Main decoder function.
 */
int main( int argc, char** argv )
{
  ERROR_CODE              err;
  ERROR_CODE_frameDeMux   err_fdm;
  ERROR_CODE_decLayer123  err_l123;
  ERROR_CODE_fileWriter   err_fw;
  
  unsigned long       gotBits = 0;
  
  unsigned int        totalErrorCount = 0;
  
  /* Frame counter */
  unsigned int        frameNum = 0;
      
  char *decPara = ""; /* @TODO */
  
  /* Init all variables, parse cmdline
   * ================================= */
  HANDLE_bitBuffer    bitstreamBuffer = NULL; /* input mpeg12 stream */
  HANDLE_bitBuffer    audioFrame      = NULL; /* logical demuxed audio frame */
  HANDLE_decLayer123  decoderCore     = NULL; /* decoder core handler */
  HANDLE_pcmBuffer    pcmBuffer       = NULL; /* output pcm buffer */
  HANDLE_frameDeMux   frameDeMux      = NULL; /* frame demuxer */
  HANDLE_fileWriter   fileWriter      = NULL; /* file writer handler */       
  HANDLE_debug        debug           = NULL; /* global debug handler */
  HANDLE_debug        debug_stream    = NULL; /* stream debug handler */
  arguments           args  = { NULL, NULL, NULL, 0, 0 };

  err = NO_ERROR;

  /* Init debug system */
  debug = debugInit(D_NOTICE, stdout);
  debugEnter(debug, "main()");
  
  printf("ISO 13818-3 MPEG-2 Reference Audio Decoder Software\n");
#ifdef USELIB_AFsp
  printf("compiled with AFsp library libtsp\n");
#else
  printf("compiled without AFsp library libtsp\n");
#endif
  
  /* Parse command line */
  if (! getArguments(argc, argv, &args, debug)) 
  {
    debugPrintf (debug, D_DEBUG, "Error while parsing command line. Terminating program.");
    goto bail;
  };

  REM( "Initialize input/output buffers" )
  bitstreamBuffer = bitBufferInit(MAX_BS_FRAME_SIZE, "bitstreamBuffer", debug);
  if ( bitstreamBuffer == NULL )
  {
    debugPrintf (debug, D_ERROR, "Could not initialize bitstream buffer.");
    err = MEMORY_ALLOCATION_ERROR;
    goto bail;
  }
  if ( !bitBufferOpenInputFile(bitstreamBuffer, args.encoded_file_name) )
  {
    debugPrintf (debug, D_ERROR, "Could not open input file: %s", args.encoded_file_name);
    goto bail;
  }

  audioFrame = bitBufferInit(MAX_AU_FRAME_SIZE, "audioFrame", debug);
  if ( audioFrame == NULL ) 
  {
    debugPrintf (debug, D_ERROR, "Could not initialize decoder input buffer.");
    err = MEMORY_ALLOCATION_ERROR;
    goto bail;
  }

  pcmBuffer = pcmBufferInit(MAX_FRAME_PCM_SAMPLES);
  if ( pcmBuffer == NULL ) 
  {
    debugPrintf (debug, D_ERROR, "Could not initialize decoder output buffer.");
    err = MEMORY_ALLOCATION_ERROR;
    goto bail;
  }

  REM( "Initialize frame demuxer" )
  frameDeMux = frameDeMuxInit(bitstreamBuffer, debug);
  if (frameDeMux == NULL)
  {
    debugPrintf(debug, D_ERROR, "Frame demuxer could not be initialized.");
    err = UNKNOWN_ERROR;
    goto bail;
  }  

  REM( "Initialize decoder core" )
  decoderCore = decLayer123Init(args.decPara, stdout);
  if ( decoderCore == NULL )
  {
    debugPrintf (debug, D_ERROR, "Could not initialize decoder core.");
    err = UNKNOWN_ERROR;
    goto bail;
  }
  
  REM( "Initialize file writer" )
  fileWriter = fileWriterInit(args.decoded_file_name, args.format, debug);


  /* Run main decoder
   * ================ */
  
  /* Extracting audioFrames from bitstream and decode them */
  while (!err)
  {
    /* Print out frame num */
    debugSetPrefix(debug, "Frame %u", frameNum); 

    /* DeMux next audioFrame
     * ===================== */
    err_fdm = frameDeMuxNext(frameDeMux, audioFrame);
    switch(err_fdm)
    {
      case FMUX_NO_ERROR:
        debugPrintf(debug, D_NOTICE_L1, "Demuxed frame corretly.");
        /* valid frame decoded - do nothing here */
        break;
        
      case FMUX_END_OF_STREAM:
        debugPrintf(debug, D_NOTICE_L1, "End of stream.");
        err = END_OF_STREAM;
        continue;
        
      case FMUX_INVALID_FRAME_LENGTH:
        debugPrintf(debug, D_WARNING, "Invalid frame length. Frame will be discarted.");
        /* no valid frame - nothing to decode - do next frame*/
        continue;

      case FMUX_NOT_ENOUGH_DATA:
        debugPrintf(debug, D_NOTICE_L1, "Not enough prebuffered data to decode. Frame will be discarted.");
        /* no valid frame - nothing to decode - do next frame*/
        continue;
      
      default:
        debugPrintf(debug, D_ERROR, "Unknown error while demuxing frame.");
        err = UNKNOWN_ERROR;
        goto bail;
    }

    /* Decode frame in mpeg layer123 library 
     * ===================================== */
    err_l123 = decLayer123Frame(decoderCore, audioFrame, pcmBuffer);
    switch(err_l123) 
    {
      case L123_NO_ERROR:  /* frame decoded correctly */
        debugPrintf(debug, D_NOTICE_L1, "Frame decoded correctly.");
        break;
        
      case L123_CRC_ERROR: /* TODO data corrected by */
        debugPrintf(debug, D_WARNING, "CRC error in frame.");
        break;
        
      case L123_UNDERRUN_ERROR:
        debugPrintf(debug, D_ERROR, "Not enough available data in demuxed frame.");
        err = UNKNOWN_ERROR;
        goto bail;
        
      default:       /* default action when error indicated */
        debugPrintf(debug, D_ERROR, "Unknown error while decoding frame.");
        err = UNKNOWN_ERROR;
        goto bail;
    }    

    /* Write out pcm buffer to file 
     * ============================ */
    err_fw = fileWriterOutBuffer(fileWriter, pcmBuffer);
    switch(err_fw)
    {
      case FW_NO_ERROR:
        debugPrintf(debug, D_NOTICE_L1, "Successfully written pcm data.");
        /* do nothing */
        break;

      default:       /* default action when error indicated */
        debugPrintf(debug, D_ERROR, "Error while writing pcm data into file.");
        err = UNKNOWN_ERROR;
        goto bail;

    }

    frameNum++;
  } /* while - for each frame! */
  
  debugSetPrefix(debug, ""); 
  debugPrintf(debug, D_NOTICE_L1, "Decoding finished.");

  /* Print some statistics
   * ===================== */
  
  {
    float gotBits         = (float) frameDeMuxGotBits(frameDeMux);
    float samplesWritten  = (float) fileWriterGetWritten(fileWriter);
    float samplingFreq    = (float) fileWriterGetSamplingFreq(fileWriter);
    float bitrate         = gotBits / samplesWritten * samplingFreq / 1000;
    float seconds         = samplesWritten / samplingFreq;
    
    debugPrintf(debug, D_NOTICE,    "Decoder framework stats:");
    debugPrintf(debug, D_NOTICE,    "  demuxed frames   :%8u frames",    frameNum);
    debugPrintf(debug, D_NOTICE_L1, "  used bits        :%8.0f bit",     gotBits);
    debugPrintf(debug, D_NOTICE,    "  written samples  :%8.0f samples", samplesWritten);
    debugPrintf(debug, D_NOTICE,    "  length           :%8.3f seconds", seconds);
    debugPrintf(debug, D_NOTICE,    "  average bitrate  :%8.3f kbps",    bitrate);
  }
  
  decLayer123PrintStats(decoderCore);
  
  
 
  /* Free all allocated data
   * ======================= */
bail:
  debugPrintf(debug, D_DEBUG_L2, "Deallocating handlers.");
  decLayer123Free(decoderCore);
  frameDeMuxFree(frameDeMux);  
  fileWriterClose(fileWriter);
  fileWriterFree(fileWriter);
  pcmBufferFree(pcmBuffer);
  bitBufferFree(audioFrame);
  bitBufferFree(bitstreamBuffer);

  debugLeave(debug);
  debugFree(debug);
  
  if (err != NO_ERROR && err != END_OF_STREAM) exit(err);
  
  return 0;

}

